'''
## This file is autogenerated from macrosynergy setup.py.

It contains version information for the macrosynergy package.

short_version: 1.2.0

version: 1.2.0

full_version: 1.2.0

git_revision: b3f53f5bc7db343d3083771e213e950374f6ad00

release: True

'''
short_version: str = '1.2.0'
version: str = '1.2.0'
full_version: str = '1.2.0'
git_revision: str = 'b3f53f5bc7db343d3083771e213e950374f6ad00'
release: bool = True
if not release:
    version = full_version
